#coding:utf-8
'''
Created on 06.04.2011

@author: kamashev
'''

import sqlite3
import json

from django.conf import settings

from datetime import datetime

from engine import BaseDataSource

from m3 import db


class SQLiteXDataSource(BaseDataSource):
    '''
    Объект, предоставляющий доступ к источнику данных формата SQLite
    '''

    def __init__(self, db_path, query, parameters = ()):
        super(SQLiteXDataSource, self).__init__()

        self.db_path = db_path
        self.query = query #точки входа параметров из кортежа параметров указывать знаками вопроса ?
        self.parameters = parameters or () #параметры, передаваемые в WHERE, HAVING, VALUES()

        self._cursor = None
        self._connection = None
        self._fetching = False
        self._fields = []

    def open_session(self):
        '''
        Открывает сессию обмена информацией с SQLite.
        '''
        self._fetching = False
        if self._cursor:
            self._cursor.close()

        self._connection = sqlite3.connect(self.db_path) #Если базы не существует?
        self._cursor = self._connection.cursor()
        self._fields = []
    
    def close_session(self):
        '''
        Закрывает сессию обмена информацией с SQLite.
        '''
        self._fetching = False
        if self._cursor:
            self._cursor.close()
            self._cursor = None
        if self._connection:
            self._connection.close()
            self._connection = None

    def read(self):
        '''
        Читает очередную порцию данных из источника данных. Возвращает None 
        в случае, если поток данных из источника закончен
        '''
        if not self._fetching:
            # выполняем запрос в базу данных
            self._cursor.execute(self.query, self.parameters)
            self._fields = map(lambda x: x[0], self._cursor.description)
            self._fetching = True
        
        for row in self._cursor:
            row_dict, i = {}, 0
            for field in self._fields:
                row_dict[field] = row[i]
                i += 1

            yield row_dict
            
#===============================================================================
# Источник данных из JSON строки с данными
#===============================================================================

class JSONDateTypeEnum(db.BaseEnumerate):
    """
    Перечисление типов объектов, которые описывают даты
    """
    DATE = 1
    DATETIME = 2

    values = {DATE:u'Дата', DATETIME:u'Дата и время'}


class JSONDataSource(BaseDataSource):
    '''
    Источник данных, который возвращает записи из 
    строки со списоком json-объектов
    '''
    def __init__(self, json_data='', date_fields=None):
        """
        @param date_fields: Словарь ключей другого словаря. Данные по этим
            ключам в другом словаре вернут объект-дату.
        @type date_fields: Объект является словарем. Его ключами являются ключи
            словаря строки данных. Его значениями являются значения перечисления
            JSONDateTypeEnum, которые описывают какой именно тип объекта-даты к
            нам поступил.
        """
        if not date_fields: date_fields = []
        super(JSONDataSource, self).__init__()
        self._json_data = json_data
        self._data = None
        self.__date_fields = date_fields
        
    def open_session(self):
        '''
        Открывает сессию взаимодействия с источником данных
        '''
        self._data = json.JSONDecoder().decode(self._json_data)
        self.__prepare_dates()
    
    def close_session(self):
        '''
        Закрывает сессию взаимодействия с источником данных
        '''
        self._data = None
    
    def read(self):
        '''
        Читаем очередную строку из источника данных
        '''
        if self._data:
            if not isinstance(self._data, list):
                if self._pos < 0:
                    self._pos += 1
                    yield self._data
            else:
                for i in xrange(0, len(self._data)):
                    yield self._data[i]

    def __prepare_dates(self):
        """
        Превращает пришедшие даты из строк в python объекты.
        """
        if self._data and self.__date_fields:
            for datum in self._data:
                for k, v in datum.iteritems():
                    date_keys = self.__date_fields.keys()
                    if k in date_keys:
                        datum[k] = self.__prepare_date_by_type(
                            v, self.__date_fields[k])

    def __prepare_date_by_type(self, datestr, type):
        """
        Преобразует строку в объект datetime либо date

        @param datestr: Строка для превращения.
        @param type: Тип объекта даты.
        """
        if type == JSONDateTypeEnum.DATE:
            return datetime.strptime(datestr, settings.DATE_FORMAT).date()
        elif type == JSONDateTypeEnum.DATETIME:
            return datetime.strptime(datestr, settings.DATE_FORMAT+' %H:%M:%S')
        